classdef TomoFilamentUIManager < handle
    % This viewer contains the following GUI types:
    %   TomoView: show the tomo volume in pseudo-3D format, and draw
    %       filaments on it.
    %   CylinView: show the unrolling surfaces and the corresponding
    %       spectrum.

    % basic setting
    properties (GetAccess = public, SetAccess = private)
        extr_samp_step
        extr_max_rad
        extr_ref_theta
        cvter_cyl2eul
        cvter_eul2cyl

        fila_trace_paras
    end

    % computation nodes
    properties (GetAccess = private, SetAccess = private)
        % inputs nodes
        ndFilaEngn

        ndOrigTomoVol
        ndTomoVol
        ndTomoLayInd

        ndTFStat

        ndStatTFFree
        ndStatTFAdd
        ndStatTFMov
        ndStatTFFoc
        ndStatTFShowCurrPt
        ndStatTFCen
        ndStatTFFTVa
        ndStatTFFTVb
        ndStatTFFTVs

        ndStatSpecRmvBg
        ndStatFlipContrast
        ndStatSurfSymmAvg

        ndFilaSecInd
        ndFilaSecAvgNum
        ndFilaInnRad
        ndFilaOutRad

        ndExtrFila
        ndExtrFilaRect
        ndExtrFilaPts
        ndExtrFilaSmthPts
        ndExtrFilaPtLens
        ndExtrFilaRMFr
        ndExtrFilaRMFs
        ndExtrFilaRMFt

        ndShowSurf

        ndSpecSegSize
        ndSpecSegStep
        ndSpecGamma
        ndSpecShowOrd
        ndSpecVa
        ndSpecVb
        ndSpecOptLen
        ndSpecOptAng

        ndTFCurrPt
        ndTFCurrPtSpc
        ndTFFocPtInd
        ndTFFocPt
        ndTFFocPtSecInd

        ndTFTracThr;

        ndTFNewFilaPts
        ndTFNewFilaPtsInSec
        ndTFNewFilaPtLens
        ndTFNewFilaSmthPts
        ndTFNewFilaSmthPtsInSec
        ndTFNewFilaSmthPtLens

        ndTomoLay
        ndCrtFilaPts
        ndSelFilaPts
        ndAllFilaPts
        ndTunFocPt

        ndCylSpec
        ndLogSpec
        ndTylLogSpec

        ndStatRpvValid

        ndHp
        ndHpSn
        ndHpRise
        ndHpTwist

        ndTFSelPtCyl

        ndTFCurrPtCds
        ndTFCurrPtEul
        ndTFCurrPtCyl

        %%% Advanced
        ndFlagFiltSurf
        ndFiltSurfAngMean
        ndFiltSurfAngRang
        ndFiltSurfThre

    end

    % const
    properties (GetAccess = private, Constant)
        default_tomoview_view_angle = [-80, 70];
        default_tomoview_layer_alpha = 0.9;
        default_tomoview_creatingfilacolor = [.8, .4, .4];
        default_tomoview_allfilacolor = [.4, .8, .4];
        default_tomoview_selectingfilacolor = [.4, .4, .8];
        default_tomoview_tuningfilacolor = [.8, .4, .4];
        default_tomoview_tuningFocusingcolor = [.4, .4, .8];
        space_euler = "Eul";
        space_cylin = "Cyl";
        tf_stat_free = "Free";
        tf_stat_adding = "Adding";
        tf_stat_moving = "Moving";
        tf_stat_focusing = "Focusing";
        tf_stat_setting_va = "SettingVa";
        tf_stat_setting_vb = "SettingVb";
        tf_stat_foc_pt_eul = "Focsing point(eul)";
        tf_stat_foc_pt_cyl = "Focsing point(cyl)";
        tf_stat_foc_ftpt = "Focsing FT point";
    end
    
    methods (Access = public)
        function obj = TomoFilamentUIManager()
            obj.extr_samp_step = 1;
            obj.extr_max_rad = 50;
            obj.extr_ref_theta = -180:1:(180-1e-6);
            obj.cvter_cyl2eul = @(cyl)([]);
            obj.cvter_eul2cyl = @(eul)([]);
            obj.buildCompGraph();
            obj.buildAdvancedCompGraph();
            fila_trace_para_list = [ ...
                "template radius (pix)", "15", ...
                "template length (pix)", "50", ...
                "tracing step (pix)", "30", ...
                "tracing point number", "10", ...
                "searching step (pix)", "1", ...
                "searching number", "3", ...
                "border reserve distance", "20", ...
            ];
            obj.fila_trace_paras = containers.Map( ...
                fila_trace_para_list(1:2:end), ...
                fila_trace_para_list(2:2:end));
        end

        buildCompGraph(obj);

        loadTomogramData(obj, tomo_volume);
        registerFilamentEditingEngine(obj, fila_engine);

        % register GUI
        registerMainFigure(obj, fig);

        registerTomoViewAxes(obj, ax);
        registerCylinAxes(obj, ax);
        registerCylinFTAxes(obj, ax);
        registerFilaSecAxes(obj, ax);

        registerLoadTomoButton(obj, butt);
        registerLoadFilaButton(obj, butt);
        registerAddFilaButton(obj, butt);
        registerAccFilaButton(obj, butt);
        registerUndFilaButton(obj, butt);
        registerClsFilaButton(obj, butt);
        registerDelFilaButton(obj, butt);
        registerShowAllFilaButton(obj, butt);
        registerSaveFilaButton(obj, butt);
        registerTuneFilaButton(obj, butt);
        registerTFNextptButton(obj, butt);
        registerTFPrevptButton(obj, butt);
        registerTFAddptButton(obj, butt);
        registerTFMovptButton(obj, butt);
        registerTFRmvptButton(obj, butt);
        registerTFConfButton(obj, butt);
        registerTFAbortButton(obj, butt);
        registerTFCancButton(obj, butt);
        registerTFDoneButton(obj, butt);
        registerTFSetvaButton(obj, butt);
        registerTFSetvbButton(obj, butt);
        registerTFOptRpvButton(obj, butt);
        registerTFReconButton(obj, butt);
        registerTFFilaAutoCenButton(obj, butt);
        registerTFEstTraceThresButton(obj, butt);
        registerTFFilaTraceButton(obj, butt);

        registerFlipContButton(obj, butt);
        registerTFSpecRmvbgButton(obj, butt);
        registerTFAvgSurfButton(obj, butt);

        registerFilaListBox(obj, list_box);

        registerNlaySpin(obj, spin);
        registerNlayLabel(obj, lab);

        registerSecLenSpin(obj, spin);
        registerSecLenLabel(obj, lab);

        registerAvgNumSpin(obj, spin);
        registerInnRadSpin(obj, spin);
        registerOutRadSpin(obj, spin);
        registerCftGammSpin(obj, spin);
        registerSegSizeSpin(obj, spin);
        registerSegStepSpin(obj, spin);
        registerMaxOrdSpin(obj, spin);
        registerTFOptLenSpin(obj, spin);
        registerTFOptAngSpin(obj, spin);

        registerTFTraceThresEdit(obj, edt);

        registerVaaEdit(obj, edt);
        registerValEdit(obj, edt);
        registerVbaEdit(obj, edt);
        registerVblEdit(obj, edt);

        registerHpcnEdit(obj, edt);
        registerHprsEdit(obj, edt);
        registerHptwEdit(obj, edt);


        registerSaveSurfButton(obj, butt);

        %%% Additional Actions
%         autoEstimateTracingThreshold(obj);
%         autoEstimateMaxMinRad(obj);

        %%% Advanced
        buildAdvancedCompGraph(obj);

        registerTFSurfFilterButton(obj, butt);
        registerTFSurfFiltAngMeanSpin(obj, spin);
        registerTFSurfFiltAngRangeSpin(obj, spin);
        registerTFSurfFiltThreSpin(obj, spin);

    end

    methods (Access = private)
        tomoViewLayerClickedCallback(obj, src, evt);
        filaSecClickedCallback(obj, src, evt);
        cylinClickedCallback(obj, src, evt);
        cylinFTClickedCallback(obj, src, evt);
        keyPressCallback(obj, src, evt);

        extrSelectingFila(obj);

        function focusOnFilaPoint(obj, foc_ind)
            sec_ind = floor(obj.ndTFNewFilaPtLens.value(foc_ind)) + 1;
            upder = obj.ndTFCurrPt.updateValue([0;0;sec_ind]);
            upder = obj.ndTFCurrPtSpc.updateValue(obj.space_cylin, upder);
            upder = obj.ndTFFocPtInd.updateValue(foc_ind, upder);
            upder = obj.ndTFStat.updateValue(obj.tf_stat_focusing, upder);
            obj.ndFilaSecInd.updateValue(sec_ind, upder);
        end

    end


end