function [trace_pos, trace_dir, trace_score, rand_score, thres_proposal] = filamentTracing3D(vol, init_pt, templ_rad,...
    templ_len, trace_step, sear_step, sear_num, varargin)

    persistent p;
    if isempty(p)
        p = inputParser();
        p.addParameter('est_rand_scores', false);
        p.addParameter('num_rand_scores', 100);
        p.addParameter('score_thres', []);
        p.addParameter('border_dist', []);
        p.addParameter('max_pt_num', 10);
        p.addParameter('arc_approx', false);
        p.addParameter('both_sides', true);
    end
    p.parse(varargin{:});

    [~, fila_axis_func, fila_axis_len, fila_pt_lens, ~] = helix.filamentSmoothing(init_pt);
    
    templ_cylin_val = extract_and_construct_template(vol, ...
        fila_axis_func, fila_axis_len, templ_len, templ_rad);
    templ_cylin_val = single(templ_cylin_val);
    
    init_dir = normalize(fnval(fila_axis_func, fila_pt_lens+0.5e-4)-fnval(fila_axis_func, fila_pt_lens-0.5e-4),1,'norm');
    init_score = random_test(vol, templ_cylin_val, size(init_dir, 2), init_pt, quat_mapztov(init_dir));

    % random test
    if p.Results.est_rand_scores
        rand_score = random_test(vol, templ_cylin_val, p.Results.num_rand_scores);
        thres_proposal = 0.5*(prctile(rand_score, 90) + min(init_score));
        fprintf("The min score of init points is %f, while the 95 prctile " + ...
            "of random test score is %f. Thus the proposed score threshold is %f.\n", ...
            min(init_score), prctile(rand_score, 90), thres_proposal);
    end

    score_thres = [];
    if ~isempty(p.Results.score_thres)
        score_thres = p.Results.score_thres;
        fprintf("The score threshold is specified as %f.\n", score_thres);
    elseif exist('thres_proposal', 'var')
        score_thres = thres_proposal;
        fprintf("The score threshold is set as %f.\n", score_thres);
    end

    for trace_side = 1:2
        if trace_side == 1
            trace_direction_tag = "Forward";
            current_pos = init_pt(:,end);
            current_dir = init_dir(:,end);
        else
            trace_direction_tag = "Backward";
            current_pos = init_pt(:,1);
            current_dir = -init_dir(:,1);
        end
        
        % start filament tracing
        trace_numb = p.Results.max_pt_num;
        one_side_trace_pos = zeros(3, trace_numb);
        one_side_trace_dir = zeros(3, trace_numb);
        one_side_trace_score = zeros(1, trace_numb);

        flag_exit = false;
        fprintf("Begin %s tracing.\n", trace_direction_tag);
        for isearch = 1:trace_numb
            fprintf("%s: determining the %d-th (of %d) points.\n", trace_direction_tag, isearch, trace_numb);
        
            %%%% Find the next point %%%%
            [next_pos, next_dir, val] = find_next_point(vol, templ_cylin_val, ...
                current_pos, current_dir, trace_step, sear_step, sear_num,...
                p.Results.arc_approx);
        
            %%%% Update results %%%%
            one_side_trace_pos(:,isearch) = next_pos;
            one_side_trace_dir(:,isearch) = next_dir;
            one_side_trace_score(isearch) = val;
    
            current_pos = next_pos;
            current_dir = next_dir;
    
            %%%% Check exit conditions %%%
            if ~isempty(score_thres)
                if val < score_thres
                    flag_exit = true;
                end
            end
            if ~isempty(p.Results.border_dist)
                bd = p.Results.border_dist(:);
                if any(next_pos < bd) || any(next_pos > size(vol)'-bd)
                    flag_exit = true;
                end
            end
    
            if flag_exit
                one_side_trace_pos = one_side_trace_pos(:, 1:isearch - 1);
                one_side_trace_dir = one_side_trace_dir(:, 1:isearch - 1);
                one_side_trace_score = one_side_trace_score(:, 1:isearch - 1);
                break;
            end
    
        end

        if trace_side == 1
            trace_pos = [init_pt, one_side_trace_pos];
            trace_dir = [init_dir, one_side_trace_dir];
            trace_score = [init_score, one_side_trace_score];
        else
            trace_pos = [one_side_trace_pos(:,end:-1:1), trace_pos]; %#ok<AGROW> 
            trace_dir = [one_side_trace_pos(:,end:-1:1), trace_dir]; %#ok<AGROW> 
            trace_score = [one_side_trace_score(end:-1:1), trace_score]; %#ok<AGROW> 
        end

    end
    
end

function [templ_cylin_val] = extract_and_construct_template(vol, fila_axis_func, fila_axis_len, templ_len, templ_rad)

    [extr_fila,~,~,~,~,~,~,~] = filamentExtraction(vol, fila_axis_func, 1, fila_axis_len, templ_rad);
    [~, ~, templ] = ring_average(sum(extr_fila,3));
    templ = templ - mean(templ,'all');
    templ_cylin = repmat(templ, [1,1,templ_len]);
    templ_sz = ceil(sqrt(sum(size(templ_cylin).^2)));
    templ_sz = templ_sz + 1 - mod(templ_sz,2);
    templ_cylin_val = cenrsz(templ_cylin, templ_sz*[1,1,1]);

end

function [next_pos, next_dir, val] = find_next_point(vol, templ_cylin_val, ...
    current_pos, current_dir, trace_step, sear_step, sear_num, ...
    flag_arc_approx)

    ref_x = [0;0;1];
    ref_y = normalize(cross(current_dir, ref_x), 'norm');
    ref_x = normalize(cross(ref_y, current_dir), 'norm');
    [sx, sy] = cencoord([sear_num,sear_num]*2+1, sear_step/trace_step);
    sx = reshape(sx, 1, []);
    sy = reshape(sy, 1, []);

    cand_dir = normalize(current_dir + (ref_x.*sx + ref_y.*sy), 1, 'norm');
    if flag_arc_approx
        cand_pos_dir = normalize(current_dir + (ref_x.*sx + ref_y.*sy), 1, 'norm');
        cand_pos = current_pos + cand_pos_dir * trace_step;
    else
        cand_pos = current_pos + cand_dir * trace_step;
    end

    cand_dir_num = size(cand_dir, 2);
    cand_pos_num = size(cand_pos, 2);

    volsz = size(templ_cylin_val);
    [vx, vy, vz] = cencoord(volsz);
    vcd = single([vx(:),vy(:),vz(:)]);
    vxb = single(cencoord(volsz(1)));
    vyb = single(cencoord(volsz(2)));
    vzb = single(cencoord(volsz(3)));

    vol_boxes = get_subvols(vol, cand_pos, volsz, 'ignore_sub_pixel', true);
    vol_boxes_linear = reshape(vol_boxes, [], size(vol_boxes,4))';
    nmlz_vol_boxes_linear = vol_boxes_linear ./ vecnorm(vol_boxes_linear, 2, 2);
    cc_score = -inf(cand_pos_num, cand_dir_num);

    for id = 1:cand_dir_num
        vcd_rt_ref = quat_mapztov(cand_dir(:,id)).rotateframe(eye(3));
        vcd_rt = vcd * vcd_rt_ref;

        %%%%========== HOTSPOT FUNCTION ==========%%%%
        rot_templ_linear = interp3(vyb, vxb, vzb, templ_cylin_val, vcd_rt(:,2), vcd_rt(:,1), vcd_rt(:,3), 'linear', 0);
        %%%%======================================%%%%

        if flag_arc_approx
            cc_score(id) = nmlz_vol_boxes_linear(id, :) * rot_templ_linear / norm(rot_templ_linear);
        else
            cc_score(:, id) = nmlz_vol_boxes_linear * rot_templ_linear / norm(rot_templ_linear);
        end

    end

    [val, idx] = max(cc_score, [], 'all');

    if flag_arc_approx
        idx_p = idx;
        idx_d = idx;
    else
        [idx_p, idx_d] = ind2sub(size(cc_score), idx);
    end

    next_dir = cand_dir(:, idx_d);
    next_pos = cand_pos(:, idx_p);

end

function rand_scores = random_test(vol, templ_cylin_val, test_num, specific_poses, specific_quats)

    volsz = size(templ_cylin_val);
    [vx, vy, vz] = cencoord(volsz);
    vcd = single([vx(:),vy(:),vz(:)]);
    vxb = single(cencoord(volsz(1)));
    vyb = single(cencoord(volsz(2)));
    vzb = single(cencoord(volsz(3)));

    if exist('specific_poses', 'var')
        rand_poses = specific_poses;
    else
        rand_poses = randn(3, test_num).*(size(vol)' - volsz') + round(volsz');
    end
    if exist('specific_quats', 'var')
        rand_qs = specific_quats;
    else
        rand_qs = randq(test_num);
    end
    
    vol_boxes = get_subvols(vol, rand_poses, volsz, 'ignore_sub_pixel', true);
    vol_boxes_linear = reshape(vol_boxes, [], size(vol_boxes,4))';
    nmlz_vol_boxes_linear = vol_boxes_linear ./ vecnorm(vol_boxes_linear, 2, 2);

    rand_scores = zeros(1, test_num);
    for itest = 1:test_num
        
        vcd_rt_ref = rand_qs(itest).rotateframe(eye(3));
        vcd_rt = vcd * vcd_rt_ref;

        rot_templ_linear = interp3(vyb, vxb, vzb, templ_cylin_val, vcd_rt(:,2), vcd_rt(:,1), vcd_rt(:,3), 'linear', 0);

        rand_scores(itest) = nmlz_vol_boxes_linear(itest, :) * rot_templ_linear / norm(rot_templ_linear);
    end
end





